#include <sta/drivers/MS56xx.hpp>

#include <sta/debug/assert.hpp>


namespace sta 
{
    MS56xx::MS56xx(SPIDevice * device, DelayUsFunc delay, OsrLevel level /* = OsrLevel::_1024 */)
        : device_{device},
          delay_{delay},
          osr_{level}
    {
        STA_ASSERT(device != nullptr);
    }

    MS56xx::MS56xx(I2CDevice * device, DelayUsFunc delay, OsrLevel osr = OsrLevel::_1024)
        : device_{device},
          delay_{delay},
          osr_{osr}
    {
        STA_ASSERT(device != nullptr);
    }

    bool MS56xx::init()
    {
        // Reset device on start-up
        this->reset();

        // Read the PROM e.g. constants
        this->readPROM();
    }

    void MS56xx::pulseCS(uint32_t ms=1) {
        this->device_->endTransmission();
        delay_(ms * 1000);
        this->device_->beginTransmission();
    }

    void MS56xx::setOsr(OsrLevel osr)
    {
        osr_ = osr;
    }

    void MS56xx::reset()
    {
        busCommand(MS56xx::Operations::RESET);
        delay_(MS56xx::RESET_DELAY);
    }

    // Helper method to keep code clean
    void MS56xx::pulseCS(uint32_t ms=1) {
        this->device_->endTransmission();
        delay_(ms * 1000);
        this->device_->beginTransmission();
    }

    int32_t MS56xx::getPressure(int32_t temperature, Unit unit /* = Unit::hPa */)
    {
        // Request the ADC to read pressure values.
        busCommand(MS56xx::Operations::D1_CONVERSION + 2*this->osr_);

        // This might be an incorrect duration.
        delay_(10 * 1000);

        uint8_t buffer[3] = { 0x00, 0x00, 0x00 };
        busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);

        uint32_t d1_val = buffer[0] << 16 | buffer[1] << 8 | buffer[2];
        int32_t pressure = calculatePressure(d1_val, reverseTempCalc(temperature), unit);

        return pressure;
    }

    int32_t MS56xx::getPressure(Unit unit = Unit::hPa)
    {
        int32_t temperature = getTemperature();

        return getPressure(temperature, unit);
    }

    int32_t MS56xx::getTemperature() {
        busCommand(MS56xx::Operations::D2_CONVERSION + 2*this->osr_);

        // This might be an incorrect duration.
        delay_(10 * 1000);

        uint8_t buffer[3] = { 0x00, 0x00, 0x00 };
        busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);

        // Convert the raw values to actual temperature values.
        uint32_t res = buffer[0] | buffer[1] << 8 | buffer[2] << 16;
        int32_t temperature = calculateTemperature(res);

        return temperature;
    }

    void MS56xx::setPressureReference(float pressRef, float altRef, Unit unit /* = Unit::hPa */)
    {
        // TODO
    }

    float MS56xx::getAltitudeEstimate(int32_t temperature)
    {
        int32_t pressure = getPressure(temperature, Unit::hPa);
    }

    float MS56xx::getAltitudeEstimate()
    {
        int32_t temperature = getTemperature();

        return getAltitudeEstimate(temperature);
    }

    int32_t MS56xx::reverseTempCalc(int32_t temp) {
        return this->tempsens;
    }

    int32_t MS56xx::calculatePressure(uint32_t d1, int32_t dT, Unit unit) {
        int64_t offset = ( ((uint64_t)this->off) << 17) + ( ( ((uint64_t)this->tco) * dT ) >> 6);
        int64_t sensitivity = ( ((uint64_t)this->sens) << 16) + ( ( ((uint64_t)this->tcs) * dT ) >> 7);
        int32_t pres = ( (( ((uint64_t)d1) * sensitivity) >> 21) - offset ) >> 15;

        switch (unit)
        {
        case Unit::hPa:
            /* code */
            break;
        case Unit::Pa:

            break;
        case Unit::mbar:
            break;

        default:
            break;
        }

        return pres;
    }

    int32_t MS56xx::calculateTemperature(uint32_t d2) {
        int32_t dT = d2 - ( ((uint32_t)this->t_ref) << 8);
        int32_t temp = 2000 + ((dT * ((uint32_t)this->tempsens)) >> 23);

        // Further calculations for low (<20) and very low (<(-15)) could be possible
        // But I don't know whether they are necessary

        return temp;
    } 

    // Read all constants from the PROM
    // May be moved to be called in reset() function in future
    // Request value x -> Read value x; Then request value y etc.
    // Could be optimized; Not as important since only needed once at start-up
    void MS56xx::readPROM() {
        this->device_->beginTransmission();

        this->device_->transfer(Operations::READ_PROM);
        uint8_t sensArr[2];
        this->device_->receive(sensArr, 2);
        this->sens = uint_8BufferTouint16_t(sensArr);

        pulseCS();

        this->device_->transfer(Operations::READ_PROM+2);
        uint8_t offArr[2];
        this->device_->receive(offArr, 2);
        this->off = uint_8BufferTouint16_t(offArr);

        pulseCS();

        this->device_->transfer(Operations::READ_PROM+4);
        uint8_t tcsArr[2];
        this->device_->receive(tcsArr, 2);
        this->sens = uint_8BufferTouint16_t(sensArr);

        pulseCS();

        this->device_->transfer(Operations::READ_PROM+6);
        uint8_t tcoArr[2];
        this->device_->receive(tcoArr, 2);
        this->tco = uint_8BufferTouint16_t(tcoArr);

        pulseCS();

        this->device_->transfer(Operations::READ_PROM+8);
        uint8_t t_refArr[2];
        this->device_->receive(t_refArr, 2);
        this->t_ref = uint_8BufferTouint16_t(t_refArr);

        pulseCS();

        this->device_->transfer(Operations::READ_PROM+0xA);
        uint8_t tempsensArr[2];
        this->device_->receive(tempsensArr, 2);
        this->tempsens = uint_8BufferTouint16_t(tempsensArr);

        this->device_->endTransmission();
    }

    // Helper function:
    // Take first bytes from buffer, swap them and store those in uint16_t
    // Swap may not be necessary
    uint16_t MS56xx::uint_8BufferTouint16_t(uint8_t* buffer) {
        return (buffer[0] << 8) | buffer[1];
    }

    bool MS56xx::busCommand(uint8_t command)
    {
        this->device_->beginTransmission();
        this->device_->transfer(command);
        this->device_->endTransmission();

        return true;
    }

    bool MS56xx::busRead(uint8_t reg, uint8_t * buffer, size_t length)
    {
        this->device_->beginTransmission();
        this->device_->transfer(reg);
        this->device_->receive(buffer, length);
        this->device_->endTransmission();

        return true;
    }

    bool MS56xx::busWrite(uint8_t reg, uint8_t * buffer, size_t length)
    {
        return true;
    }
}